package flowserver

import (
	"encoding/json"
	"errors"
	"flowserver/msg"
	"log"
	"net/http"
	"sync"

	"github.com/gorilla/websocket"
)

//FlowSessionManager or FlowServerCore
type FlowSessionManager struct {
	// List of flow sessions?
	sessions map[string]*FlowSession

	sync.Mutex
}

//NewFlowSessionManager creates a New initialized FlowSessionManager
func NewFlowSessionManager() *FlowSessionManager {
	return &FlowSessionManager{
		sessions: map[string]*FlowSession{},
	}
}

//CreateSession creates a new session
func (fsm *FlowSessionManager) CreateSession() *FlowSession {
	fsm.Lock()
	defer fsm.Unlock()
	for {
		ID := RandString(10)
		sess, ok := fsm.sessions[ID]
		if !ok {
			sess = NewSession(ID)
			fsm.sessions[ID] = sess
			return sess
		}
	}
}

//LoadSession loads or creates a new Session
func (fsm *FlowSessionManager) LoadSession(ID string) (*FlowSession, error) {
	fsm.Lock()
	defer fsm.Unlock()
	if ID == "" {
		return nil, errors.New("Cannot be null")
	}
	sess, ok := fsm.sessions[ID]
	if !ok {
		sess = NewSession(ID)
		fsm.sessions[ID] = sess
	}
	return sess, nil

}

var upgrader = websocket.Upgrader{}

func (fsm *FlowSessionManager) ServeHTTP(w http.ResponseWriter, r *http.Request) {

	c, err := upgrader.Upgrade(w, r, nil)
	if err != nil {
		log.Println("upgrade:", err)
		return
	}
	defer c.Close()

	// Room
	var sess *FlowSession
	defer func() {
		if sess == nil {
			return
		}
		// Remove client on exit
		sess.ClientRemove(c)
	}()

	/////////////////
	// Message handling
	// ////////

	// Websocket IO loop
	for {
		mt, data, err := c.ReadMessage()
		if err != nil {
			log.Println("Err:", err)
			break
		}
		if mt != websocket.TextMessage {
			log.Println("Not a text message?")
			break
		}

		fmsg := msg.RecvMessage{}
		err = json.Unmarshal(data, &fmsg)
		if err != nil {
			log.Println("Err parsing message:", err)
			c.WriteJSON("bye")
			break
		}
		switch fmsg.OP {
		/////////////////////////////
		// NEWSESSION request
		//////////////////
		case "sessionNew":
			err = func() error {
				log.Println("We want a new session so")
				sess = fsm.CreateSession()
				sess.ClientAdd(c)
				return c.WriteJSON(msg.SendMessage{OP: "sessionSwitch", Data: sess.ID})
			}()
			//////////////////////////////////
			// LOADSESSION request
			////////////////////////////////////
		case "sessionLoad":
			err = func() error {
				sessID := string(fmsg.Data)
				if sess != nil {
					sess.ClientRemove(c)
				}
				sess, err = fsm.LoadSession(sessID) // Set our session
				if err != nil {
					log.Println("Err:", err)
					return err
				}
				return sess.ClientAdd(c)
			}()
		///////////////////////
		// DOCUMENTUPDATE Receive a document
		//////
		case "documentUpdate":
			err = func() error {
				if sess == nil {
					return errors.New("There is no session")
				}
				return sess.DocumentUpdate(c, fmsg.Data)
			}()
		}

		if err != nil {
			log.Println("Err Writing", err)
			break
		}
	}
	log.Println("ws Is disconnecting")
}
